import pytest
import numpy as np

from sklearn.datasets import fetch_20newsgroups
from numpy.testing import (
    assert_almost_equal,
    assert_array_equal,
    assert_allclose,
)
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.naive_bayes import MultinomialNB

from uncwrap import UncertaintyWrapperEstimator
from uncwrap.tests.test_utils import load_politics_data, PoliticsClassifierWrapper


@pytest.fixture
def newsgroups_data():
    newsgroups_train = fetch_20newsgroups(subset='train')
    return newsgroups_train.data, newsgroups_train.target

@pytest.fixture
def politics_data():
    return load_politics_data("./uncwrap/tests/fixtures/bbc-text.csv")


def test_pretrained_classifier(newsgroups_data, politics_data):
    # Given
    news_X, news_y = newsgroups_data
    polit_X_train, polit_X_test, polit_y_train, polit_y_test = politics_data
    tfidf_vectorizer = TfidfVectorizer(max_features=1000)
    tfidf_vectorizer.fit(news_X)

    #   We build a very simple naive bayes model using the 20 newsgroups data that will become the blackbox model
    blackbox_model = MultinomialNB()
    blackbox_model.fit(tfidf_vectorizer.transform(news_X), news_y)

    #   we create the target classifier
    politics_model = PoliticsClassifierWrapper(blackbox_model)

    # When

    #   and use the uncertainty wrapper to check it
    uncertainty_wrapper = UncertaintyWrapperEstimator(black_box=politics_model)
    uncertainty_wrapper.fit(tfidf_vectorizer.transform(polit_X_train), polit_y_train)

    polit_y_pred_uncert = uncertainty_wrapper.predict(tfidf_vectorizer.transform(polit_X_test))

    # Then
    assert_almost_equal(polit_y_pred_uncert[:, 1].max(), 0.693, decimal=3)
    assert_almost_equal(polit_y_pred_uncert[:, 1].min(), 0.000, decimal=3)
    assert_array_equal(politics_model.predict(tfidf_vectorizer.transform(polit_X_test)), polit_y_pred_uncert[:, 0])


def test_no_pretrained_classifier_and_no_preds(newsgroups_data, politics_data):
    # Given
    news_X, news_y = newsgroups_data
    polit_X_train, polit_X_test, polit_y_train, polit_y_test = politics_data
    tfidf_vectorizer = TfidfVectorizer(max_features=1000)
    tfidf_vectorizer.fit(news_X)

    #   we create the target classifier
    politics_model = None

    # When

    #   and use the uncertainty wrapper to check it
    uncertainty_wrapper = UncertaintyWrapperEstimator(black_box=None)
    # Then
    with pytest.raises(ValueError, match="No blackbox or predictions was set"):
        uncertainty_wrapper.fit(tfidf_vectorizer.transform(polit_X_train), polit_y_train)


def test_pretrained_classifier_wrong_input_shapes(newsgroups_data, politics_data):
    # Given
    news_X, news_y = newsgroups_data
    polit_X_train, polit_X_test, polit_y_train, polit_y_test = politics_data
    tfidf_vectorizer = TfidfVectorizer(max_features=1000)
    tfidf_vectorizer.fit(news_X)

    #   we create the target classifier
    politics_model = None

    # When
    X = tfidf_vectorizer.transform(polit_X_train)
    y = news_y
    #   and use the uncertainty wrapper to check it
    uncertainty_wrapper = UncertaintyWrapperEstimator(black_box=None)
    # Then
    with pytest.raises(ValueError, match=r'Found input variables with inconsistent numbers of samples: [[\d, \d]'):
        uncertainty_wrapper.fit(X, y)


def test_pretrained_classifier_with_preds(newsgroups_data, politics_data):
    # Given
    news_X, news_y = newsgroups_data
    polit_X_train, polit_X_test, polit_y_train, polit_y_test = politics_data
    tfidf_vectorizer = TfidfVectorizer(max_features=1000)
    tfidf_vectorizer.fit(news_X)

    #   We build a very simple naive bayes model using the 20 newsgroups data that will become the blackbox model
    blackbox_model = MultinomialNB()
    blackbox_model.fit(tfidf_vectorizer.transform(news_X), news_y)

    #   we create the target classifier
    politics_model = PoliticsClassifierWrapper(blackbox_model)

    # When
    X = tfidf_vectorizer.transform(polit_X_train)
    y = polit_y_train
    pred_y = politics_model.predict(X)

    #   and use the uncertainty wrapper to check it
    uncertainty_wrapper = UncertaintyWrapperEstimator()
    uncertainty_wrapper.fit(X, y, pred_y)

    X_test = tfidf_vectorizer.transform(polit_X_test)
    pred_y_test = politics_model.predict(X_test)
    polit_y_pred_uncert = uncertainty_wrapper.predict(X_test, pred_y_test)

    # Then
    assert_almost_equal(polit_y_pred_uncert[:, 1].max(), 0.693, decimal=3)
    assert_almost_equal(polit_y_pred_uncert[:, 1].min(), 0.000, decimal=3)
    assert_array_equal(pred_y_test, polit_y_pred_uncert[:, 0])


import numpy as np
from sklearn.datasets import fetch_20newsgroups
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.naive_bayes import MultinomialNB
from sklearn.metrics import accuracy_score
from uncwrap import UncertaintyWrapperEstimator
from uncwrap.tests.test_utils import PoliticsClassifierWrapper, load_politics_data


def main():
    """

    This script shows how to apply the uncertainty wrapper to enrich the predictions of a trained classifier with
    the corresponding uncertainty when applied to a new domain of application. The goal is to show how to use
    a classifier already trained for one task in another task, and how we can measure that the blackbox adapts for the
    new task. We will see how wecan identify the cases where the blackbox is less confident on its predictions,
    and how we can increase the accuracy of the resulting system by rejecting those examples that more uncertain.

    It first trains a naive bayes classifier over newsgroups that will be the blackbox in this example.
    Then, it uses it for predicting whether a new text is talking about politics or not, by grouping all the
    newsgroups that are related to politics.
    As the target application, it uses the dataset on BBC news(https://www.kaggle.com/yufengdev/bbc-fulltext-and-category),
    making a binary classifier to predict if a news talks about politics or not.
    """

    # First we load the 20 newsgroups dataset
    newsgroups_train = fetch_20newsgroups(subset='train')
    newsgroups_test = fetch_20newsgroups(subset='test')

    X_train = newsgroups_train.data
    X_test = newsgroups_test.data

    y_train = newsgroups_train.target
    y_test = newsgroups_test.target

    # We create a tf/idf vectorizer for texts. It is important to remark here that it will be the same that we use
    # for computing the uncertainty, as we need to use the same representation for obtaining the predictions when
    # calling the blackbox trained for newsgroups.
    tfidf_vectorizer = TfidfVectorizer(max_features=1000)
    tfidf_vectorizer.fit(X_train)

    # We build a very simple naive bayes model using the 20 newsgroups data that will become the blackbox model
    blackbox_model = MultinomialNB()
    blackbox_model.fit(tfidf_vectorizer.transform(X_train), y_train)

    # We compute the accuracy of the 20 newsgroups dataset to check that the blackbox works
    y_pred = blackbox_model.predict(tfidf_vectorizer.transform(X_test))
    print('accuracy of the black-box in test: {}'.format(accuracy_score(y_test, y_pred)))

    # To apply the blackbox for the binary problem of predicting if a news talks about politics or not
    # we create a wrapper that returns 1 for those news that are related to politics and 0 otherwise
    politics_model = PoliticsClassifierWrapper(blackbox_model)
    y_pred2 = politics_model.predict(tfidf_vectorizer.transform(X_test))

    # Load the BBC news dataset also making a transformation for getting 1 when the category is politics
    # and 0 otherwise
    polit_X_train, polit_X_test, polit_y_train, polit_y_test = load_politics_data("..//tests/fixtures/bbc-text.csv")

    # we create the uncertaintywrapper referencing to the black box
    uncertainty_wrapper = UncertaintyWrapperEstimator(black_box=politics_model, verbose=1)

    # and train the wrapper (mind that we are using the same vectorizer for the input as in the blackbox)
    uncertainty_wrapper.fit(tfidf_vectorizer.transform(polit_X_train), polit_y_train)

    # then we can obtain pairs of predictions and uncertainties
    polit_y_pred_uncert = uncertainty_wrapper.predict(tfidf_vectorizer.transform(polit_X_test))

    # and compare the accuracy when removing the predictions that are more uncertain in the new problem
    polit_y_pred = polit_y_pred_uncert[:, 0]
    polit_y_uncerts = polit_y_pred_uncert[:, 1]
    print('accuracy of the black-box for politics: {}'.format(accuracy_score(polit_y_test, polit_y_pred)))
    UNCERTAINTY_THRESHOLD = 0.69
    polit_y_uncertain_preds_indexes = polit_y_uncerts < UNCERTAINTY_THRESHOLD
    print('accuracy of the black-box for politics for keeping the more {} confident preds out of {}: {}'.format(
        np.count_nonzero(polit_y_uncertain_preds_indexes),
        polit_y_uncerts.shape[0],
        accuracy_score(np.array(polit_y_test)[polit_y_uncertain_preds_indexes],
                       polit_y_pred[polit_y_uncertain_preds_indexes])))

if __name__ == '__main__':  # pragma: no cover
    main()


"""
This is a module to be used as a reference for building other modules
"""
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels
from tensorflow.keras import backend as K
from tensorflow.keras.utils import to_categorical

from uncwrap.models.dirichlet_wrapper import DirichletUncertaintyWrapper


class UncertaintyWrapperEstimator(ClassifierMixin, BaseEstimator):
    """ An Sklearn wrapper that enriches the prediction of a given classifier with the corresponding prediction ucnertainty

    For more information look
    at the paper:
    Mena, J., Pujol, O., & Vitrià, J. (2020).
    Uncertainty-based Rejection Wrappers for Black-box Classifiers. IEEE Access.

    Parameters
    ----------
    black_box : BaseEstimator, default=None
        An object that observes the sklearn BaseEstimator interface. This will be used as the black box to retrieve the original predictions. Optional, if not set, then a list of predictions on X must be passed on fit and predict |
    lambda_reg : float, default=1e-2
        Regularization property used for adjusting the uncertainties.
    num_samples : int, default=1000
        Number of samples to include for the MC sampling of the dirichlet dsitribution.
    learning_rate : float, default='demo'
        Learning rate used when training uncertainties.
    num_hidden_units : int, default='demo'
        Number of units to use in the uncertainty neural network model.
    batch_size : int, default='demo'
        Batch size used when training uncertainties.
    epochs : int, default='demo'
        Number of epochs when training the uncertainties.
    verbose : int, default='demo'
        Verbosity level ofthe training(same as in Keras).

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`fit`.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    wrapper_ : object
        The model used for learning and estimating the uncertainties
    """

    def __init__(self, black_box: BaseEstimator = None, lambda_reg=1e-2, num_samples=1000,
                 learning_rate=1e-3, num_hidden_units=20, batch_size=256, epochs=50, verbose=0):
        self.black_box = black_box
        self.lambda_reg = lambda_reg
        self.num_samples = num_samples
        self.learning_rate = learning_rate
        self.num_hidden_units = num_hidden_units
        self.batch_size = batch_size
        self.epochs = epochs
        self.verbose = verbose

    def fit(self, X, y, pred_y = None):
        """A reference implementation of a fitting function for a classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.
        pred_y : array-like, shape (n_samples,)
            Array of predictions corresponding to calls to the blackbox for X.
            It can be used when no access to the representation of the input for the blackbox.
            An array of int.

        Returns
        -------
        self : object
            Returns self.
        """
        # Check that X and y have correct shape
        X, y = check_X_y(X, y, accept_sparse='csr')
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        if not self.black_box and not type(pred_y) is np.ndarray:
            raise ValueError("No blackbox or predictions was set")

        self.X_ = X
        self.y_ = y
        if not type(pred_y) is np.ndarray:
            pred_y = self.black_box.predict(X)
        if X.getformat() == 'csr':
            X = X.toarray()
        if len(y.shape) == 1:
            y = to_categorical(y, len(self.classes_))
        if len(pred_y.shape) == 1:
            pred_y = to_categorical(pred_y, len(self.classes_))
        self.wrapper_ = DirichletUncertaintyWrapper(lambda_reg=self.lambda_reg,
                                                    num_samples=self.num_samples,
                                                    learning_rate=self.learning_rate,
                                                    num_hidden_units=self.num_hidden_units, verbose=self.verbose)
        self.wrapper_.train_model(X, y, pred_y, batch_size=self.batch_size, epochs=self.epochs)
        # Return the classifier
        return self

    def predict(self, X, pred_y = None):
        """ A reference implementation of a prediction for a classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        pred_y : array-like, shape (n_samples,)
            Array of predictions corresponding to calls to the blackbox for X.
            It can be used when no access to the representation of the input for the blackbox.
            An array of int.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The original prediction together with the corresponding uncertainty for each prediction.
        """
        # Input validation
        X = check_array(X, accept_sparse=True)

        if not type(pred_y) is np.ndarray:
            pred_y = self.black_box.predict(X)
        if X.getformat() == 'csr':
            X = X.toarray()
        if len(pred_y.shape) == 1:
            pred_y = to_categorical(pred_y, len(self.classes_))
        uncertainties = K.get_session().run(self.wrapper_.predict_entropy(X=X, pred_y=pred_y))
        result = np.array((np.argmax(pred_y, axis=1), uncertainties))
        return result.T



from tensorflow.keras.layers import Dense, Input, concatenate
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.metrics import categorical_accuracy
import tensorflow as tf
import tensorflow_probability as tfp


class DirichletUncertaintyWrapper(object):
    """ A wrapper that estimates the uncertainty of the predictions of a classifier by modeling each point
        as a Dirichlet distribution.

        For more information look at the paper:
        Mena, J., Pujol, O., & Vitrià, J. (2020).
        Uncertainty-based Rejection Wrappers for Black-box Classifiers. IEEE Access.

        Parameters
        ----------
        lambda_reg : float, default=1e-2
            Regularization property used for adjusting the uncertainties.
        num_samples : int, default=1000
            Number of samples to include for the MC sampling of the dirichlet dsitribution.
        learning_rate : float, default='demo'
            Learning rate used when training uncertainties.
        num_hidden_units : int, default='demo'
            Number of units to use in the uncertainty neural network model.
        batch_size : int, default='demo'
            Batch size used when training uncertainties.
        epochs : int, default='demo'
            Number of epochs when training the uncertainties.
        verbose : int, default='demo'
            Verbosity level ofthe training(same as in Keras).

        Attributes
        ----------
        num_classes_ : int
            The number of classes to predict.
        model_ : object
            The model used for learning and estimating the uncertainties
        training_history_: object
            Training history
        """
    def __init__(self, lambda_reg=1e-2, num_samples=1000,
                 learning_rate=1e-3, num_hidden_units=20,
                 verbose=0):
        self.lambda_reg = lambda_reg
        self.epsilon = 1e-10
        self.num_samples = num_samples
        self.learning_rate = learning_rate
        self.num_hidden_units = num_hidden_units
        self.verbose = verbose

    def dirichlet_aleatoric_cross_entropy(self, y_true, y_pred):
        """
            Loss function that applies a categorical cross entropy to the predictions
            obtained from the original model combined with a beta parameter that models
            the aleatoric noise associated with each data point. We model a Dirichlet
            pdf based on the combination of both the prediction and the beta parameter
            to sample from it and obtain the resulting output probabilities

            Parameters
            ----------
            y_true:  `np.array`
                the labels in one hot encoding format

            y_pred: `np.array`
                output of the model formed by the concatenation of the original prediction
                in the first num_classes positions, and a beta scalar in the last one


            Returns
            -------
            an array with the associated cross-entropy for each element of the batch

        """
        # original predictions obtained with the blackbox
        mu_probs = y_pred[:, :self.num_classes_]
        # scale param estimated by the wrapper
        beta = y_pred[:, self.num_classes_:]
        # parameter of the Dirichlet distribution
        alpha = mu_probs * beta
        dirichlet = tfp.distributions.Dirichlet(alpha)
        # sample the Dir distribution for each point
        z = dirichlet.sample(sample_shape=self.num_samples)
        # compute the crossentropy for the mean of the samples
        e_probs = tf.reduce_mean(z, axis=0)
        log_probs = tf.log(e_probs + self.epsilon)
        cross_entropy = -(tf.reduce_sum(y_true * log_probs, axis=-1))
        # the loss includes a regularization term for constraining the beta to grow unbounded
        return cross_entropy + self.lambda_reg * tf.reduce_sum(beta, axis=-1)

    def max_beta(self, y_true, y_pred, **args):
        """
            metric that outputs the max value for the beta

            Parameters
            ----------
            y_true:  `np.array`
                the labels in one hot encoding format

            y_pred: `np.array`
                output of the model formed by the concatenation of the original prediction
                in the first num_classes positions, and a beta scalar in the last one


            Returns
            -------
            the max value for the beta

        """
        beta = y_pred[:, self.num_classes_:]
        return tf.reduce_max(beta)

    def min_beta(self, y_true, y_pred, **args):
        """
            metric that outputs the min value for the beta

            Parameters
            ----------
            y_true:  `np.array`
                the labels in one hot encoding format

            y_pred: `np.array`
                output of the model formed by the concatenation of the original prediction
                in the first num_classes positions, and a beta scalar in the last one


            Returns
            -------
            the min value for the beta

        """
        beta = y_pred[:, self.num_classes_:]
        return tf.reduce_min(beta)

    # metric that outputs the accuracy when only considering the logits_mu.
    # this accuracy should be the same that was obtained with the fake classifier
    # in its best epoch.
    # def mu_accuracy(self):
    #  num_classes = self.num_classes
    def mu_accuracy(self, y_true, y_pred, **args):
        """
            metric that outputs the accuracy when only considering the logits_mu.
            This accuracy should be the same that was obtained with the fake classifier in its best epoch.

            Parameters
            ----------
            y_true:  `np.array`
                the labels in one hot encoding format

            y_pred: `np.array`
                output of the model formed by the concatenation of the original prediction
                in the first num_classes positions, and a beta scalar in the last one


            Returns
            -------
            the accuracy of the original predictions

        """
        mu_probs = y_pred[:, :self.num_classes_]
        y_true_probs = y_true[:, :self.num_classes_]
        return categorical_accuracy(y_true_probs, mu_probs)

    def create_model(self, input_shape):
        """
            Method that creates the Keras Neural Network model for the DL  wrapper that will estimate the betas.

            Parameters
            ----------
            input_shape:  `int`
                shape of the input


            Returns
            -------
            a model to be trained

        """
        model_input = Input(shape=(input_shape,))
        logits_sigma = Dense(self.num_hidden_units, activation='relu')(model_input)
        logits_sigma = Dense(self.num_hidden_units, activation='relu')(logits_sigma)
        logits_sigma = Dense(self.num_hidden_units, activation='relu')(logits_sigma)
        logits_sigma = Dense(self.num_hidden_units, activation='relu')(logits_sigma)
        logits_sigma = Dense(1, activation='sigmoid')(logits_sigma)
        probs_mu = Input(shape=(self.num_classes_,))
        output = concatenate([probs_mu, logits_sigma])

        model = Model(inputs=[model_input, probs_mu], outputs=output)
        model.compile(loss=self.dirichlet_aleatoric_cross_entropy,
                      optimizer=Adam(lr=self.learning_rate),
                      metrics=[self.mu_accuracy, self.min_beta, self.max_beta])
        return model

    def train_model(self, X, y, pred_y, batch_size=256, epochs=50):
        """
            Method that trains the wrapper for the inputs, and the corresponding labels and predictions of the blackbox.

            Parameters
            ----------
            X : array-like, shape (n_samples, n_features)
            The training input samples.
            y : array-like, shape (n_samples,)
                The target values. An array of int.
            pred_y : array-like, shape (n_samples,)
                Array of predictions corresponding to calls to the blackbox for X.
                It can be used when no access to the representation of the input for the blackbox.
                An array of int.
            batch_size : int, default='demo'
                Batch size used when training uncertainties.
            epochs : int, default='demo'
                Number of epochs when training the uncertainties.


            Returns
            -------
            None

        """
        self.num_classes_ = pred_y.shape[1]
        input_shape = X.shape[1]
        self.model_ = self.create_model(input_shape)
        self.training_history_ = self.model_.fit([X, pred_y],
                                                 y,
                                                 batch_size=batch_size,
                                                 epochs=epochs,
                                                 shuffle=True,
                                                 verbose=self.verbose,
                                                 validation_split=0.2)

    def predict_entropy(self, X, pred_y):
        """
            Method that trains the wrapper for the inputs, and the corresponding labels and predictions of the blackbox.

            Parameters
            ----------
            X : array-like, shape (n_samples, n_features)
                The training input samples.
            pred_y : array-like, shape (n_samples,)
                Array of predictions corresponding to calls to the blackbox for X.
                It can be used when no access to the representation of the input for the blackbox.
                An array of int.


            Returns
            -------
            An array with the uncertainty for each data input and prediction

        """
        y_pred = self.model_.predict([X, pred_y])
        # original predictions obtained with the blackbox
        mu_probs = y_pred[:, :self.num_classes_]
        # scale param estimated by the wrapper
        beta = y_pred[:, self.num_classes_:]
        # parameter of the Dirichlet distribution
        alpha = mu_probs * beta
        dirichlet = tfp.distributions.Dirichlet(alpha)
        # sample the Dir distribution for each point
        z = dirichlet.sample(sample_shape=self.num_samples)
        # compute the entropy between the original pred and the mean of the samples
        e_probs = tf.reduce_mean(z, axis=0)
        log_probs = tf.log(e_probs + self.epsilon)
        entropy = tf.reduce_sum(-e_probs * log_probs, axis=-1)
        return entropy



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uncwrap', 'uncwrap.examples', 'uncwrap.models', 'uncwrap.tests']

package_data = \
{'': ['*'], 'uncwrap.tests': ['fixtures/*']}

install_requires = \
['click==7.1.2',
 'minilog==1.6',
 'numpy==1.19.2',
 'pandas==1.1.3',
 'protobuf>=3.19.0,<4.0.0',
 'scikit-learn==0.23.2',
 'scipy==1.5.2',
 'tensorflow-probability==0.7.0',
 'tensorflow==1.15.2']

entry_points = \
{'console_scripts': ['UncertaintyModeler = '
                     'uncertainty_classification_wrapper.cli:main']}

setup_kwargs = {
    'name': 'uncwrap',
    'version': '0.5',
    'description': 'Sklearn wrapper for classifications system that computes the prediction uncertainty',
    'long_description': '# Uncertainty assessment for Pre-trained models\n\nJose Mena\n\n![](RackMultipart20200930-4-16wcrro_html_f8a36c35410c1486.jpg)\n\n## Motivation\n\nImagine one of the following situations:\n\n- You have trained a model for medical diagnosis to enable the early detection of an illness based on images obtained from an X-ray machine of a hospital. You train a model using labelled data to predict whether the patient has the illness or not based on her corresponding image. Now, a new hospital is interested on your model and they ask you to start using it, but then you realize that they use a different X-ray machine. To which extent is your model good at predicting using the new images? How can you trust these predictions?\n- You find an API that offers a Sentiment Analysis classifier. At the documentation you can read that the model was trained using reviews of Amazon products. The point is that you want to apply it to reviews of restaurants that you plan to use as implicit feedback for a tourism recommender system. How will the API work in the new domain? A horror movie can be terrifying and thus have a good review, but I don&#39;t see how a terrifying restaurant could be good.\n\nOf course, there might be situations where getting some labelled data for the target domain can enable to fine-tune the original model or even training it from scratch, but sometimes this is not an option, as we don&#39;t have access to the original model or to the resources needed for training it.\n\nWhat if we could have a way to tell if the prediction for the new dataset is confident or not? Picture a system that, together with the prediction of the model, you could have a metric of its uncertainty. Usually, a useful way of measuring this uncertainty, in probabilistic classifiers, is by observing the output probabilities. A probability of 0.9 for the predicted class would mean a very confident prediction. The problem with this approach is that there is no guarantee that these probabilities are well-calibrated (that they are real probabilities), and, besides, there might be cases where the prediction might be very confident, like this 0.9, but that it corresponds to a wrong prediction.\n\nTake for example, a linear classifier trained with the date at the top-left figure. When we apply to the target data, bottom-left figure, we observe that there is a mass of points that will be wrongly classified, as they are in the class 1 region, but they are labelled as class 2.\n\n![](uncertainties.png)\n\nImagine that this classifier was trained to predict if an image belongs to a cat, a dog or a bear. Imagine that the images used for training that classifier included white cats, but only black or brown bears. At prediction time, we get pictures of polar bears. Our classifier we&#39;ll try to do its best and it may predict them as cats, as it only had seen white animals belonging to the cat category.\n\n![](cat.jpg) ![](bear.jpeg)\n\n## Uncertainty wrapper\n\nThe component here presented tackles the scenario of having a pre-trained classifier, from now on a black-box, and trying to apply it to a new target scenario.\n\nWhat we propose in this component is a wrapper that takes the input and the output of the black-box and enriches the later with a measure of uncertainty. Based on this uncertainty, you are be free of &quot;trusting&quot; the prediction or discard it. If the overall results of applying the black box to your problem are poor, maybe you would have to reconsider the option of labelling more data and train a specific classifier for the target problem.\n\nFurther theoretical details can be found at the paper [https://ieeexplore.ieee.org/document/9097854](https://ieeexplore.ieee.org/document/9097854), but the main idea is that we have a black-box classification system that takes as input a feature vector **x** , or an embedding, in the case of images and texts, and outputs a probability distribution **y\\*** over the predicted categories or classes. The wrapper produces a new prediction **y&#39;** together with an uncertainty score for this prediction **u** , as illustrated in the following diagram:\n\n![](diagram2.png)\n\nIn the instantiation of the wrapper, these are the parameters that can be set:\n\n| Name | Type | Mandatory |\n| --- | --- | --- |\n| **black\\_box** | An object that observes the sklearn BaseEstimator interface. This will be used as the black box to retrieve the original predictions | Optional, if not set, then a list of predictions on X must be passed on fit and predict |\n| **lambda_reg** | Regularization property used for adjusting the uncertainties | no, default 1e-2\n| **num_samples** | Number of samples to include for the MC sampling of the dirichlet dsitribution | no, default 1000 |\n| **learning_rate** | Learning rate used when training uncertainties | no, default 1e-3 |\n| **num_hidden_units** | Number of units to use in the uncertainty neural network model | no, default 20 |\n| **batch_size** | Batch size used when training uncertainties | no, default 256 |\n| **epochs** | Number of epochs when training the uncertainties | no, default 50 |\n| **verbose** | Verbosity level ofthe training(same as in Keras) | no, default 0 |\n\nUsually these parameters produces good estimates for the uncertainty but, depending on the case, they have to be tuned to avoid overfitting.\n\nThe wrapper offers two methods (following the sklearn standard):\n\n### fit\n\nThis method is meant for training the wrapper. It needs some examples of the target domain **(x, y)** and will call the black-box for the corresponding **y\\***.\n\n| Name | Type | Mandatory |\n| --- | --- | --- |\n| **X** | Vector of input vectors | yes |\n| **y** | Vector of labels | yes |\n| **pred_y** | Vector of blackbox predictions on X | yes, when there is no blackbox defined |\n\n#### Returns\n\nA wrapper model trained for the target dataset\n\n### Predict\n\nThis method applies the wrapper to obtain the prediction and the associated uncertainty. It takes the input **x** and calls the black-box for a list of the corresponding **y\\***.\n\n| Name | Type | Mandatory |\n| --- | --- | --- |\n| **X** | Vector of input vectors | yes |\n| **pred_y** | Vector of blackbox predictions on X | yes, when there is no blackbox defined |\n\n#### Returns\n\nA list of pairs (y\\*, u) corresponding to the prediction and the degree of uncertainty.\n\n_IMPORTANT: Note that the output of the black box, the original prediction can be in the form of a probability distribution over the categories of the classifier, as resulting from the usual softmax layer, or a categorical output, consisting on the identifier of the predicted class. In both cases, the wrapper will be able to add the corresponding uncertainty._\n\n## Example\n\nIn the examples folder you will find an example(script or notebook) of a potential use case for the library.\nThe example shows how to apply the uncertainty wrapper to enrich the predictions of a trained classifier with \nthe corresponding uncertainty when applied to a new domain of application. The goal is to show how to use\na classifier already trained for one task in another task, and how we can measure that the blackbox adapts for the \nnew task. We will see how wecan identify the cases where the blackbox is less confident on its predictions, \nand how we can increase the accuracy of the resulting system by rejecting those examples that more uncertain.\n\n# Setup\n\n## Requirements\n\n* Python 3.6+\n* scipy\n* pandas\n* numpy\n* scikit-learn\n* tensorflow==1.15.2\n* tensorflow_probability==0.7.0\n\n## Installation\n\nInstall it directly into an activated virtual environment:\n\n```text\n$ pip3 install uncwrap\n```\n\n# Usage\n\nAfter installation, the package can be imported:\n\n```text\n$ python\n>>> import uncwrap\n>>> uncwrap.__version__\n```\n',
    'author': 'José Mena',
    'author_email': 'menajosep@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/menajosep/uncertainty_modeller',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""A sample CLI."""

import click
import log

from . import UncertaintyWrapperEstimator


@click.command()
@click.argument('feet')
def main(feet=None):
    log.init()

    wrapper = UncertaintyWrapperEstimator()


if __name__ == '__main__':  # pragma: no cover
    main()



import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split


def load_politics_data(data_path):
    data = pd.read_csv(data_path)

    def process_label(row):
        if row == 'politics':
            return 1
        else:
            return 0

    data['label'] = data['category'].apply(process_label)
    train_data, test_data = train_test_split(data, test_size=0.2)
    return list(train_data['text'].values), list(test_data['text'].values), \
           list(train_data['label'].values), list(test_data['label'].values)


class PoliticsClassifierWrapper(BaseEstimator):
    def __init__(self, black_box: BaseEstimator = None):
        self.black_box = black_box

    def predict(self, X):
        orig_preds = self.black_box.predict(X)
        # we select the cats belonging to politics
        politics_categories = [16, 17, 18, 19]
        y_pred = np.array([int(pred in politics_categories) for pred in orig_preds])
        return y_pred